from typing import List

from .base import ConceptType
from .self import ConceptSelfType


class ConceptUnionType(ConceptType):
    def __init__(self, inners: List[ConceptType]):
        self.inners = inners

    def simplify(self) -> ConceptType:
        seen = set()
        result = set()
        for inner in [inner.simplify() for inner in self.inners]:
            if inner not in seen:
                seen.add(inner)
                result.add(inner)

        if len(result) == 1:
            return next(iter(self.inners)).simplify()

        simplified_union = ConceptUnionType(list(result))

        inner_unions = simplified_union.find(ConceptUnionType)
        if inner_unions:
            for inner_union in inner_unions:
                new_simplified_union = simplified_union.replace(inner_union, ConceptSelfType())
                if new_simplified_union == inner_union:
                    simplified_union = new_simplified_union

        return simplified_union

    def __repr__(self) -> str:
        return f"Union[{','.join([inner.__repr__() for inner in self.inners])}]"

    def __str__(self) -> str:
        return " or ".join([f"`{inner.__str__()}`" for inner in list(self.inners)])

    def children(self) -> List[ConceptType]:
        return self.inners

    def replace(self, old: ConceptType, new: ConceptType) -> ConceptType:
        if self == old:
            return new
        replaced_inners = [inner.replace(old, new) for inner in self.inners]
        return ConceptUnionType(replaced_inners)

    def __eq__(self, other) -> bool:
        if isinstance(other, ConceptUnionType):
            return set(self.inners) == set(other.inners)
        return False

    def __hash__(self) -> int:
        combined_hash = hash("Union")
        for inner in self.inners:
            combined_hash ^= hash(inner)
        return combined_hash



from .errors import NotFoundError, TypeMismatchError
from .filter import (FilterBinaryExpression, FilterBinaryOperator,
                     FilterExpression, FilterExpressionVisitor,
                     FilterUnaryExpression, FilterUnaryOperator,
                     NounPhrasesExpression, ValueExpression)
from .noun_phrase import NounPhrase
from .table import Table



import functools
import inspect
from types import FunctionType
from typing import Optional

from ..docstring import DocstringParser
from ..reflection import CredentialDescriptor
from ..reflection.factory import BookCustomAuthenticationFactory
from ..reflection.types import CredentialScalarType


def connect(*args, **kwargs):
    fn: Optional[FunctionType] = None
    noun_phrase: Optional[str] = None
    name: Optional[str] = None

    if len(args) == 1 and not kwargs and inspect.isfunction(args[0]):
        fn = args[0]
        noun_phrase = None
        name = None
    else:
        fn = None
        noun_phrase = kwargs.get("noun_phrase", None)
        name = kwargs.get("name", None)

    def decorator(fn):
        if not inspect.isfunction(fn):
            raise TypeError("The connect decorator can only be applied to functions.")

        short_description = None
        signature = inspect.signature(fn)
        credentials = []

        if not fn.__doc__:
            raise ValueError("missing docstring")

        parsed_docstring = DocstringParser.parse(fn.__doc__)

        if parsed_docstring.short_description:
            short_description = parsed_docstring.short_description.strip()

        for param_name, parameter in signature.parameters.items():
            if param_name == "self":
                continue

            if parameter.annotation is not str:
                raise TypeError(f"parameter '{param_name}' is not a supported type")

            if parsed_docstring:
                param_docstring = parsed_docstring.param_by_name(param_name)
            else:
                param_docstring = None

            credential = CredentialDescriptor(
                id=param_name,
                label=param_docstring.label if param_docstring is not None else None,
                type=CredentialScalarType.TEXT,
                description=(param_docstring.description if param_docstring is not None else None),
            )

            credentials.append(credential)

        if not hasattr(fn, "__connect__"):
            fn.__connect__ = BookCustomAuthenticationFactory.create(id=fn.__name__, noun_phrase_str=noun_phrase, credentials=credentials, description=short_description, name=name)

        if not hasattr(fn, "__signature__"):
            fn.__signature__ = signature

        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            return fn(*args, **kwargs)

        return wrapper

    if fn is None:
        return decorator

    return decorator(fn)



from __future__ import annotations

import inspect
from dataclasses import dataclass, field
from inspect import Signature
from typing import List, Optional

from .book_procedure_signature import BookProcedureSignature
from .concept_descriptor import ConceptDescriptor
from .example import Example
from .parameter_concept_bind import ParameterConceptBind


@dataclass
class BookProcedureDescriptor:
    id: str
    english_signature: BookProcedureSignature
    python_signature: Signature
    parameter_concept_map: List[ParameterConceptBind]
    examples: List[Example] = field(default_factory=list)
    filter_capable: bool = False
    page_capable: bool = False
    connection_required: bool = False
    override_connection_required: Optional[bool] = None
    outputs: Optional[List[ConceptDescriptor]] = None
    short_description: Optional[str] = None
    long_description: Optional[str] = None
    filter_argument_position: Optional[int] = None
    offset_argument_position: Optional[int] = None
    limit_argument_position: Optional[int] = None

    @property
    def input_concepts(self) -> Optional[List[ConceptDescriptor]]:
        input_concepts = []

        for parameter_concept in self.parameter_concept_map:
            input_concepts.extend(parameter_concept.concepts)

        return input_concepts if len(input_concepts) != 0 else None

    @property
    def output_concepts(self) -> Optional[List[ConceptDescriptor]]:
        return self.outputs

    @property
    def filter_argument_name(self) -> Optional[str]:
        return self.argument_name(self.filter_argument_position)

    @property
    def offset_argument_name(self) -> Optional[str]:
        return self.argument_name(self.offset_argument_position)

    @property
    def limit_argument_name(self) -> Optional[str]:
        return self.argument_name(self.limit_argument_position)

    @property
    def filter_has_default(self) -> Optional[bool]:
        return self.has_default(self.filter_argument_position)

    @property
    def offset_has_default(self) -> Optional[bool]:
        return self.has_default(self.offset_argument_position)

    @property
    def limit_has_default(self) -> Optional[bool]:
        return self.has_default(self.limit_argument_position)

    def argument_name(self, pos: Optional[int]) -> Optional[str]:
        if pos is None:
            return None
        parameters = list(self.python_signature.parameters.keys())
        if len(parameters) > pos:
            return parameters[pos]
        return None

    def has_default(self, pos: Optional[int]) -> Optional[bool]:
        if pos is None:
            return None

        parameters = list(self.python_signature.parameters.values())

        if 0 <= pos < len(parameters):
            return parameters[pos].default is not inspect.Parameter.empty

        return None
